package net.sourceforge.jetris;

/* MainFrame created on 14.09.2006 */
import javax.imageio.ImageIO;
import javax.swing.*;
import javax.swing.border.BevelBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.text.html.HTML;

import res.ResClass;

import net.sourceforge.jetris.io.PublishHiScore;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowFocusListener;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

public class JetrisMainFrame extends JFrame {

	private static final String NAME = "JETRIS 1.1";
	private static final int CELL_H = 24;

	private Font font;
	private JPanel playPanel;
	private JLabel score;
	private JLabel lines;
	private JLabel time;
	private JLabel[] statsF;
	private JLabel[] statsL;
	private JLabel levelLabel;
	private JLabel hiScoreLabel;
	private JPanel[][] cells;
	private TetrominoGrid tg;
	private JPanel[][] next;
	private int nextX;
	private int nextY;
	private Figure f;
	private Figure fNext;
	private FigureFactory ff;
	private boolean isNewFigureDroped;
	private boolean isGameOver;
	private boolean isPause;
	private Color nextBg = new Color(238, 238, 238);
	private Color emptyCell = new Color(150, 150, 150);
	private Color goalUndefCell = new Color(0, 0, 255);
	private Color goalFullCell = new Color(150, 255, 85);

	private TimeThread tt;
	private KeyListener keyHandler;
	private Iterator<Action> actionIter;
	private boolean playSolution=false;
	
	private List<Level> llist = null;
	private int currentLevel = 0;
	
	private JPanel about;
	private JPanel main;

	// MENU
	private JMenuItem jetrisRestart;
	private JMenuItem jetrisPause;
	private JMenuItem jetrisHiScore;
	private JMenuItem jetrisExit;

	private JMenuItem helpAbout;
	private JMenuItem helpJetris;

	private HelpDialog helpDialog;

	private JPanel hiScorePanel;
	private PublishHandler pH;

	private class GridThread extends Thread {

		private int count = 0;

		public void run() {
			try {
				while (true) {
					if (isGameOver || isPause) {
						Thread.sleep(50);
					} else {
						if (isNewFigureDroped) {
							isNewFigureDroped = false;
							count = 0;
							nextMove();
							continue;
						} else {
							Thread.sleep(50);
						}
						count += 50;
						if (count + 50 * tg.getLevel() >= 50) {
							count = 0;
							if (playSolution && !isPause) 
								nextAction();
						}
					}
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	private class TimeThread extends Thread {

		private int hours;
		private int min;
		private int sec;

		private int count;

		private void incSec() {
			sec++;
			if (sec == 60) {
				sec = 0;
				min++;
			}
			if (min == 60) {
				min = 0;
				hours++;
			}
		}

		private void resetTime() {
			hours = min = sec = 0;
		}

		public void run() {
			try {
				while (true) {
					Thread.sleep(50);
					if (isGameOver) {
						Graphics g = playPanel.getGraphics();
						Font font = new Font(g.getFont().getFontName(),
								Font.BOLD, 24);
						g.setFont(font);
						g.drawString("GAME OVER", 47, 250);

					} else if (isPause) {
						time.setText("PAUSED");
						Graphics g = playPanel.getGraphics();
						if (g!=null) {
							Font font = new Font(g.getFont().getFontName(),
									Font.BOLD, 24);
							g.setFont(font);
							g.drawString("PAUSED", 47, 250);
						}
					} else if (count >= 1000) {
						count = 0;
						incSec();
						time.setText(this.toString());
					} else {
						count += 50;
					}
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		public String toString() {
			StringBuffer sb = new StringBuffer();
			if (hours < 10) {
				sb.append('0');
			}
			sb.append(hours);

			sb.append(':');

			if (min < 10) {
				sb.append('0');
			}
			sb.append(min);

			sb.append(':');

			if (sec < 10) {
				sb.append('0');
			}
			sb.append(sec);

			return sb.toString();
		}
	}

	public JetrisMainFrame() {
		super(NAME);

		setIconImage(loadImage("jetris16x16.png"));

		keyHandler = new KeyAdapter() {

			public void keyPressed(KeyEvent e) {
				int code = e.getKeyCode();
				if (code == KeyEvent.VK_A || code == KeyEvent.VK_LEFT) {
					moveLeft();
				} else if (code == KeyEvent.VK_D || code == KeyEvent.VK_RIGHT) {
					moveRight();
				} else if (code == KeyEvent.VK_S || code == KeyEvent.VK_DOWN) {
					moveDown();
				} else if (code == KeyEvent.VK_W || code == KeyEvent.VK_UP) {
					rotation();
				} else if (code == KeyEvent.VK_SPACE) {
					moveDrop();
				} /*
				 * else if(code == KeyEvent.VK_R) { //Only for the applet needed
				 * restart(); } else if(code == KeyEvent.VK_P) { pause(); }
				 */
			}
		};
		addKeyListener(keyHandler);

		pH = new PublishHandler();

		font = new Font("Dialog", Font.PLAIN, 12);
		
		//---- Parse and setup level. ------------
		
		LevelParser lp;		
		
		try {
			lp = new LevelParser("levels.txt");
			LevelParser.levels();
			llist = lp.getLevels();
		} catch (FileNotFoundException e) {
			System.out.println("Could not find levels.txt");
		} catch (ParseException e) {
			System.out.println(e.toString());
		}

		initMenu();
		main = new JPanel(new BorderLayout());

		GridThread gt = new GridThread();
		tt = new TimeThread();

		//start with first level read from file
		setupLevel(llist.get(currentLevel));
		exchangePlayPanel();
		
		main.add(getStatPanel(), BorderLayout.WEST);
		main.add(getMenuPanel(), BorderLayout.EAST);
		main.add(getCopyrightPanel(), BorderLayout.SOUTH);
		
		
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		this.getContentPane().add(main, BorderLayout.CENTER);
		pack();
		this.setResizable(false);
		
		fNext = ff.getNextFromSeq();
		dropNext();
		paintTG();
		nextMove();
		
		addWindowFocusListener(new WindowFocusListener() {

			public void windowGainedFocus(WindowEvent arg0) {
			}

			public void windowLostFocus(WindowEvent arg0) {
				isPause = true;
			}
		});

		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		setLocation(screenSize.width / 2 - getWidth() / 2, screenSize.height
				/ 2 - getHeight() / 2);
		setVisible(true);
		
		gt.start();
		tt.start();
	}

	private void exchangePlayPanel() {
		if (playPanel != null) main.remove(playPanel);
		playPanel = getPlayPanel();
		main.add(playPanel, BorderLayout.CENTER);
		pack();
		paintTG();
	}

	private void setupLevel(Level lvl) {
		tg = new TetrominoGrid(lvl);
		ff = new FigureFactory(lvl.getObjSeq());
	}
	
	private void solveLevel(Level lvl) {
		if (lvl.getSolution().size() == 0) {
			Solver s = new Solver(lvl);
			lvl.setSolution(Solver.solutionKeys);
		}
		actionIter = lvl.getSolution().iterator();
	}
	
	void previousLevel() {
		if (currentLevel > 0) {
			currentLevel--;
			setupLevel(llist.get(currentLevel));
			exchangePlayPanel();
			restart();
		}		
	}
	
	void nextLevel() {
		if (currentLevel < (llist.size()-1)) {
			currentLevel++;
			setupLevel(llist.get(currentLevel));
			exchangePlayPanel();
			restart();
		}
	}
	


	private void initMenu() {

		MenuHandler mH = new MenuHandler();

		JMenuBar menu = new JMenuBar();
		setJMenuBar(menu);

		JMenu mJetris = new JMenu();
		menu.add(mJetris);
		mJetris.setText("Jetris");
		mJetris.setMnemonic('J');
		{
			jetrisRestart = new JMenuItem("Restart");
			mJetris.add(jetrisRestart);
			setKeyAcceleratorMenu(jetrisRestart, 'R', 0);
			jetrisRestart.addActionListener(mH);
			jetrisRestart.setMnemonic('R');

			jetrisPause = new JMenuItem("Pause");
			mJetris.add(jetrisPause);
			setKeyAcceleratorMenu(jetrisPause, 'P', 0);
			jetrisPause.addActionListener(mH);
			jetrisPause.setMnemonic('P');

			mJetris.addSeparator();

			jetrisHiScore = new JMenuItem("HiScore...");
			mJetris.add(jetrisHiScore);
			setKeyAcceleratorMenu(jetrisHiScore, 'H', 0);
			jetrisHiScore.addActionListener(mH);
			jetrisHiScore.setMnemonic('H');

			mJetris.addSeparator();

			jetrisExit = new JMenuItem("Exit");
			mJetris.add(jetrisExit);
			setKeyAcceleratorMenu(jetrisExit, KeyEvent.VK_ESCAPE, 0);
			jetrisExit.addActionListener(mH);
			jetrisExit.setMnemonic('X');
		}

		JMenu mHelp = new JMenu();
		menu.add(mHelp);
		mHelp.setText("Help");
		mHelp.setMnemonic('H');
		{
			helpJetris = new JMenuItem("Jetris Help");
			mHelp.add(helpJetris);
			setKeyAcceleratorMenu(helpJetris, KeyEvent.VK_F1, 0);
			helpJetris.addActionListener(mH);
			helpJetris.setMnemonic('J');

			helpAbout = new JMenuItem("About");
			mHelp.add(helpAbout);
			helpAbout.addActionListener(mH);
			helpAbout.setMnemonic('A');
		}
	}

	private void setKeyAcceleratorMenu(JMenuItem mi, int keyCode, int mask) {
		KeyStroke ks = KeyStroke.getKeyStroke(keyCode, mask);
		mi.setAccelerator(ks);
	}

	private JPanel getPlayPanel() {
		playPanel = new JPanel();
		playPanel.setLayout(new GridLayout(tg.getHeight(), tg.getWidth()));
		playPanel.setPreferredSize(new Dimension(tg.getWidth() * CELL_H, tg.getHeight() * CELL_H));

		cells = new JPanel[tg.getHeight()][tg.getWidth()];
		for (int i = 0; i < tg.getHeight(); i++) {
			for (int j = 0; j < tg.getWidth(); j++) {
				cells[i][j] = new JPanel();
				cells[i][j].setBackground(Color.WHITE);
				paintBorder(j,i);
				playPanel.add(cells[i][j]);
			}
		}
		return playPanel;
	}

	private JPanel getMenuPanel() {
		JPanel r = new JPanel();
		BoxLayout rL = new BoxLayout(r, BoxLayout.Y_AXIS);
		r.setLayout(rL);
		r.setBorder(new EtchedBorder());
		Dimension ra = new Dimension(5, 0);
		next = new JPanel[4][4];
		JPanel nextP = new JPanel();
		nextP.setLayout(new GridLayout(4, 4));
		Dimension d = new Dimension(4 * 18, 4 * 18);
		nextP.setMinimumSize(d);
		nextP.setPreferredSize(d);
		nextP.setMaximumSize(d);
		for (int i = 0; i < 4; i++) {
			for (int j = 0; j < 4; j++) {
				next[i][j] = new JPanel();
				nextP.add(next[i][j]);
			}
		}

		JPanel jp = new JPanel();
		jp.setLayout(new BoxLayout(jp, BoxLayout.LINE_AXIS));
		jp.add(Box.createRigidArea(ra));
		jp.add(new JLabel("NEXT:"));
		jp.add(Box.createHorizontalGlue());
		r.add(jp);
		r.add(nextP);

		r.add(Box.createRigidArea(new Dimension(100, 10)));

		jp = new JPanel();
		jp.setLayout(new BoxLayout(jp, BoxLayout.LINE_AXIS));
		jp.add(Box.createRigidArea(ra));
		jp.add(new JLabel("HI-SCORE:"));
		jp.add(Box.createHorizontalGlue());
		r.add(jp);

		hiScoreLabel = new JLabel("" + tg.hiScore[0].score);
		hiScoreLabel.setForeground(Color.RED);

		jp = new JPanel();
		jp.setLayout(new BoxLayout(jp, BoxLayout.LINE_AXIS));
		jp.add(Box.createRigidArea(ra));
		jp.add(hiScoreLabel);
		jp.add(Box.createHorizontalGlue());
		r.add(jp);

		r.add(Box.createVerticalStrut(5));

		jp = new JPanel();
		jp.setLayout(new BoxLayout(jp, BoxLayout.LINE_AXIS));
		jp.add(Box.createRigidArea(ra));
		jp.add(new JLabel("SCORE:"));
		jp.add(Box.createHorizontalGlue());
		r.add(jp);

		score = new JLabel("0");
		score.setForeground(Color.BLUE);

		jp = new JPanel();
		jp.setLayout(new BoxLayout(jp, BoxLayout.LINE_AXIS));
		jp.add(Box.createRigidArea(ra));
		jp.add(score);
		jp.add(Box.createHorizontalGlue());
		r.add(jp);

		jp = new JPanel();
		jp.setLayout(new BoxLayout(jp, BoxLayout.LINE_AXIS));
		jp.add(Box.createRigidArea(ra));
		jp.add(new JLabel("LINES:"));
		jp.add(Box.createHorizontalGlue());
		r.add(jp);

		lines = new JLabel("0");
		lines.setForeground(Color.BLUE);

		jp = new JPanel();
		jp.setLayout(new BoxLayout(jp, BoxLayout.LINE_AXIS));
		jp.add(Box.createRigidArea(ra));
		jp.add(lines);
		jp.add(Box.createHorizontalGlue());
		r.add(jp);

		jp = new JPanel();
		jp.setLayout(new BoxLayout(jp, BoxLayout.LINE_AXIS));
		jp.add(Box.createRigidArea(ra));
		jp.add(new JLabel("LEVEL:"));
		jp.add(Box.createHorizontalGlue());
		r.add(jp);

		levelLabel = new JLabel("1");
		levelLabel.setForeground(Color.BLUE);

		jp = new JPanel();
		jp.setLayout(new BoxLayout(jp, BoxLayout.LINE_AXIS));
		jp.add(Box.createRigidArea(ra));
		jp.add(levelLabel);
		jp.add(Box.createHorizontalGlue());
		r.add(jp);

		jp = new JPanel();
		jp.setLayout(new BoxLayout(jp, BoxLayout.LINE_AXIS));
		jp.add(Box.createRigidArea(ra));
		jp.add(new JLabel("TIME:"));
		jp.add(Box.createHorizontalGlue());
		r.add(jp);

		time = new JLabel("00:00:00");
		time.setForeground(Color.BLUE);

		jp = new JPanel();
		jp.setLayout(new BoxLayout(jp, BoxLayout.LINE_AXIS));
		jp.add(Box.createRigidArea(ra));
		jp.add(time);
		jp.add(Box.createHorizontalGlue());
		r.add(jp);

		r.add(Box.createVerticalGlue());

		r.add(addHelpPanel("A or \u2190 - Left"));
		r.add(addHelpPanel("D or \u2192 - Right"));
		r.add(addHelpPanel("W or \u2191 - Rotate"));
		r.add(addHelpPanel("S or \u2193 - Down"));
		r.add(addHelpPanel("Space - Drop"));

		// BUTTONS
		r.add(Box.createRigidArea(new Dimension(0, 10)));

		jp = new JPanel();
		jp.setLayout(new BoxLayout(jp, BoxLayout.LINE_AXIS));
		jp.add(Box.createRigidArea(ra));
		JButton restartBut = new JButton("Restart");
		restartBut.setToolTipText("Press 'R'");
		restartBut.setFocusable(false);
		restartBut.addKeyListener(keyHandler);
		restartBut.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				restart();
			}
		});
		d = new Dimension(120, 30);
		restartBut.setMinimumSize(d);
		restartBut.setPreferredSize(d);
		restartBut.setMaximumSize(d);
		jp.add(restartBut);
		jp.add(Box.createHorizontalGlue());
		r.add(jp);

		r.add(Box.createRigidArea(new Dimension(0, 5)));

		jp = new JPanel();
		jp.setLayout(new BoxLayout(jp, BoxLayout.LINE_AXIS));
		jp.add(Box.createRigidArea(ra));
		JButton pauseBut = new JButton("Pause");
		pauseBut.setToolTipText("Press 'P'");
		pauseBut.setFocusable(false);
		pauseBut.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				pause();
			}
		});
		pauseBut.setMinimumSize(d);
		pauseBut.setPreferredSize(d);
		pauseBut.setMaximumSize(d);
		jp.add(pauseBut);
		jp.add(Box.createHorizontalGlue());
		
		r.add(jp);

		jp = new JPanel();
		JButton solveBut = new JButton("Solution");
		solveBut.setToolTipText("Press 'S'");
		solveBut.setFocusable(false);
		solveBut.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				solveLevel(llist.get(currentLevel));
				restart();
				playSolution=true;
			}
		});
		solveBut.setMinimumSize(d);
		solveBut.setPreferredSize(d);
		solveBut.setMaximumSize(d);
		jp.add(solveBut);
		jp.add(Box.createHorizontalGlue());
		
		r.add(jp);
		
		d = new Dimension(50, 30);
		
		jp = new JPanel();
		JButton prevLvl = new JButton("<");
		prevLvl.setFocusable(false);
		prevLvl.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				previousLevel();
			}
		});
		prevLvl.setMinimumSize(d);
		prevLvl.setPreferredSize(d);
		prevLvl.setMaximumSize(d);
		jp.add(prevLvl);
		jp.add(Box.createHorizontalGlue());
		JButton nextLvl = new JButton(">");
		nextLvl.setFocusable(false);
		nextLvl.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				nextLevel();
			}
		});
		nextLvl.setMinimumSize(d);
		nextLvl.setPreferredSize(d);
		nextLvl.setMaximumSize(d);
		jp.add(nextLvl);
		jp.add(Box.createHorizontalGlue());
		
		
		r.add(jp);

		return r;
	}

	private JPanel addHelpPanel(String help) {
		JPanel jp = new JPanel();
		jp.setLayout(new BoxLayout(jp, BoxLayout.LINE_AXIS));
		jp.add(Box.createRigidArea(new Dimension(5, 0)));
		JLabel jL = new JLabel(help);
		jL.setFont(font);
		jL.setForeground(Color.GRAY);
		jp.add(jL);
		jp.add(Box.createHorizontalGlue());
		return jp;
	}

	private JPanel getCopyrightPanel() {
		JPanel r = new JPanel(new BorderLayout());
		BoxLayout rL = new BoxLayout(r, BoxLayout.X_AXIS);
		r.setLayout(rL);
		r.setBorder(new EtchedBorder());
		r.add(Box.createRigidArea(new Dimension(32, 0)));

		JLabel jL = new JLabel("Copyright (c) 2006 Nikolay G. Georgiev ");
		jL.setFont(font);
		HTMLLink email = new HTMLLink("ngg@users.sourceforge.net", true);
		email.setFont(font);

		r.add(jL);
		r.add(email);

		return r;
	}

	private JPanel getStatPanel() {
		int h = 12;
		JPanel r = new JPanel();
		BoxLayout rL = new BoxLayout(r, BoxLayout.Y_AXIS);
		r.setLayout(rL);
		r.setBorder(new EtchedBorder());

		JPanel[][] fig;
		JPanel figP, statFP;
		Dimension d = new Dimension(4 * h, 4 * h);
		Figure f;
		statsF = new JLabel[7];
		statsL = new JLabel[4];

		JPanel jp = new JPanel();
		jp.setLayout(new BoxLayout(jp, BoxLayout.LINE_AXIS));
		jp.add(Box.createRigidArea(new Dimension(5, 0)));
		jp.add(new JLabel("STATISTICS: "));
		jp.add(Box.createHorizontalGlue());
		r.add(jp);

		r.add(Box.createRigidArea(new Dimension(0, 5)));

		for (int k = 0; k < 7; k++) {
			fig = new JPanel[4][4];
			figP = new JPanel();
			statFP = new JPanel();
			statFP.setLayout(new BoxLayout(statFP, BoxLayout.LINE_AXIS));
			figP.setLayout(new GridLayout(4, 4));
			figP.setMinimumSize(d);
			figP.setPreferredSize(d);
			figP.setMaximumSize(d);

			for (int i = 0; i < 4; i++) {
				for (int j = 0; j < 4; j++) {
					fig[i][j] = new JPanel();
					fig[i][j].setBackground(nextBg);
					figP.add(fig[i][j]);
				}
			}

			switch (k + 1) {
			case Figure.I:
				f = new FigureI();
				f.setOffset(2, 0);
				break;
			case Figure.T:
				f = new FigureT();
				f.setOffset(1, 1);
				break;
			case Figure.O:
				f = new FigureO();
				f.setOffset(1, 1);
				break;
			case Figure.J:
				f = new FigureJ();
				f.setOffset(1, 1);
				break;
			case Figure.L:
				f = new FigureL();
				f.setOffset(1, 1);
				break;
			case Figure.S:
				f = new FigureS();
				f.setOffset(1, 1);
				break;
			default:
				f = new FigureZ();
				f.setOffset(1, 1);
				break;
			}

			for (int i = 0; i < 4; i++) {
				fig[f.arrY[i] + f.offsetY][f.arrX[i] + f.offsetX]
						.setBackground(f.getGolor());
				fig[f.arrY[i] + f.offsetY][f.arrX[i] + f.offsetX]
						.setBorder(BorderFactory
								.createBevelBorder(BevelBorder.RAISED));
			}
			statFP.add(figP);
			statFP.add(new JLabel("  X  "));

			statsF[k] = new JLabel("0");
			statsF[k].setForeground(Color.BLUE);
			statFP.add(statsF[k]);
			r.add(statFP);
		}

		r.add(Box.createRigidArea(new Dimension(100, 15)));

		for (int i = 0; i < statsL.length; i++) {
			statFP = new JPanel();
			statFP.setLayout(new BoxLayout(statFP, BoxLayout.LINE_AXIS));
			switch (i) {
			case 0:
				statFP.add(new JLabel("  Single  X  "));
				break;
			case 1:
				statFP.add(new JLabel("Double  X  "));
				break;
			case 2:
				statFP.add(new JLabel("  Triple  X  "));
				break;
			default:
				statFP.add(new JLabel("  Tetris  X  "));
				break;
			}
			statsL[i] = new JLabel("0");
			statsL[i].setForeground(Color.BLUE);
			statFP.add(statsL[i]);
			r.add(statFP);
			r.add(Box.createRigidArea(new Dimension(0, 5)));
		}
		return r;
	}

	static Image loadImage(String imageName) {
		try {
			Image im = ImageIO.read(new BufferedInputStream(new ResClass()
					.getClass().getResourceAsStream(imageName)));
			return im;
		} catch (Exception e) {
			e.printStackTrace(System.out);
			return null;
		}
	}

	private synchronized void nextMove() {
		if (isGameOver) return;
		f.setOffset(nextX, nextY);

		if (tg.addFigure(f)) {
			dropNext();
			if (isGameOver) return;
			f.setOffset(nextX, nextY);
			paintTG();
		} else {
			clearOldPosition();
		}
		paintNewPosition();

		if (isGameOver) {
			int tmp = tg.updateHiScore();
			if (tmp >= 0) {

				String s;

				do {
					s = JOptionPane
							.showInputDialog(
									this,
									"Enter Your Name...\nMust be between 1 and 10 charachters long",
									"New HiScore " + (tmp + 1) + ". Place",
									JOptionPane.PLAIN_MESSAGE);
				} while (s != null && (s.length() < 1 || s.length() > 10));

				if (s == null) {
					s = "<empty>";
				}

				tg.saveHiScore(s, tmp);

				if (tmp == 0)
					hiScoreLabel.setText("" + tg.hiScore[0].score);
			}
		}
	}


	private void paintBorder(int x, int y) {
		int type = tg.goalLines.get(y)[x];
		Color c = goalUndefCell;
		
		if (type == 1)
			c = goalFullCell;
		else if (type == 2)
			c = emptyCell;
		
		cells[y][x]
				.setBorder(BorderFactory.createLineBorder(c));
	}
	
	private void clearOldPosition() {
		for (int j = 0; j < 4; j++) {
			JPanel p = cells[f.arrY[j] + f.offsetYLast][f.arrX[j] + f.offsetXLast];
					
			if (tg.goalLines.get(f.arrY[j] + f.offsetYLast)[f.arrX[j] + f.offsetXLast] == 2)
					p.setBackground(Color.lightGray);
			else 
			   p.setBackground(Color.WHITE);
			paintBorder(f.arrX[j] + f.offsetXLast, f.arrY[j] + f.offsetYLast);
		}
	}

	private void paintNewPosition() {
		for (int j = 0; j < 4; j++) {
			cells[f.arrY[j] + f.offsetY][f.arrX[j] + f.offsetX].setBackground(f
					.getGolor());
			cells[f.arrY[j] + f.offsetY][f.arrX[j] + f.offsetX]
					.setBorder(BorderFactory
							.createBevelBorder(BevelBorder.RAISED));
		}
	}

	private void paintTG() {
		int i = 0;
		Color c;
		for (int[] arr : tg.gLines) {
			for (int j = 0; j < arr.length; j++) {
				if (arr[j] != 0) {
					switch (arr[j]) {
					case Figure.I:
						c = Figure.COL_I;
						break;
					case Figure.T:
						c = Figure.COL_T;
						break;
					case Figure.O:
						c = Figure.COL_O;
						break;
					case Figure.J:
						c = Figure.COL_J;
						break;
					case Figure.L:
						c = Figure.COL_L;
						break;
					case Figure.S:
						c = Figure.COL_S;
						break;
					default:
						c = Figure.COL_Z;
						break;
					}
					cells[i][j].setBackground(c);
					cells[i][j].setBorder(BorderFactory
							.createBevelBorder(BevelBorder.RAISED));
				} else {
					cells[i][j].setBackground(Color.WHITE);
					paintBorder(j,i);
				}
				if (tg.goalLines.get(i)[j] == 2)
					cells[i][j].setBackground(Color.lightGray);
			}
			i++;
		}
	}

	private void nextAction() {
		Action a;
		
		if (actionIter.hasNext()) {
			a = actionIter.next();
		} else {
			playSolution = false;
			return;
		}
		if (a==null) {
			playSolution = false;
			return;
		}
				
		switch (a) {
					case DOWN:
						moveDown();
						break;
					case LEFT:
						moveLeft();
						break;
					case RIGHT:
						moveRight();
						break;
					case ROTATE:
						rotation();
						break;
					}
	}
	
	private void showNext(Figure f) {
		for (int i = 0; i < 4; i++) {
			for (int j = 0; j < 4; j++) {
				next[i][j].setBackground(nextBg);
				next[i][j].setBorder(BorderFactory.createEmptyBorder());
			}
		}

		if (f == null)
			return;
		
		for (int j = 0; j < f.arrX.length; j++) {
			next[f.arrY[j]][f.arrX[j]].setBackground(f.getGolor());
			next[f.arrY[j]][f.arrX[j]].setBorder(BorderFactory
					.createBevelBorder(BevelBorder.RAISED));
		}
	}

	private void dropNext() {
		if (fNext == null) {
			isGameOver = true;
			return;
		}
		
		if (isGameOver)
			return;
		nextX = tg.getWidth()/2;
		nextY = 0;

		score.setText("" + tg.getScore());
		lines.setText("" + tg.getLines());
		levelLabel.setText(tg.getLevel() + " / 20");

		f = fNext;
		//fNext = ff.getRandomFigure();
		fNext = ff.getNextFromSeq();
			
		showNext(fNext);

		isGameOver = tg.isGameOver(f);

		isNewFigureDroped = true;
		updateStats();
	}

	private int moveLeft() {
		if (isGameOver || isPause)
			return 1;
		if (nextX - 1 >= 0) {
			if (tg.isNextMoveValid(f, f.offsetX - 1, f.offsetY)) {
				nextX--;
				nextMove();
			}
		}
		return 0;
	}
	
	private int moveRight() {
		if (isGameOver || isPause)
			return 1;
		if (f.getMaxRightOffset() + 1 < tg.getWidth()) {
			if (tg.isNextMoveValid(f, f.offsetX + 1, f.offsetY)) {
				nextX++;
				nextMove();
			}
		}
		return 0;
	}

	private synchronized int moveDown() {
		if (isGameOver || isPause)
			return 1;
		
		nextY++;
		nextMove();
		return 0;
	}

	private synchronized void moveDrop() {
		if (isGameOver || isPause)
			return;

		f.offsetYLast = f.offsetY;
		f.offsetXLast = f.offsetX;
		clearOldPosition();

		while (tg.isNextMoveValid(f, f.offsetX, f.offsetY)) {
			f.setOffset(f.offsetX, f.offsetY + 1);
		}

		tg.addFigure(f);
		paintTG();
		dropNext();
		nextMove();
	}

	private synchronized int rotation() {
		if (isGameOver || isPause)
			return 1;
		for (int j = 0; j < f.arrX.length; j++) {
			cells[f.arrY[j] + f.offsetY][f.arrX[j] + f.offsetX]
					.setBackground(Color.WHITE);
			paintBorder(f.arrX[j] + f.offsetX, f.arrY[j] + f.offsetY);
		}
		f.rotationRight();
		if (!tg.isNextMoveValid(f, f.offsetX, f.offsetY)) {
			f.rotationLeft();
		}
		nextMove();
		return 0;
	}

	private synchronized void pause() {
		isPause = !isPause;
		paintTG();
		paintNewPosition();
	}

	private void restart() {
		tg.resetGrid();
		paintTG();
		ff.reset();
		isGameOver = false;
		isPause = false;
		fNext = ff.getNextFromSeq();
		tt.resetTime();
		time.setText("00:00:00");
		tg.resetStats();
		
		List<Action> script = llist.get(currentLevel).getSolution(); 
		actionIter = script.iterator();
		
		playSolution = false;
		dropNext();
		nextMove();
	}

	private void updateStats() {
		for (int i = 0; i < statsF.length; i++) {
			statsF[i].setText("" + ff.getCounts()[i]);
		}

		for (int i = 0; i < statsL.length; i++) {
			statsL[i].setText("" + tg.getDropLines()[i]);
		}
	}

	private void doHelp() {
		if (helpDialog == null)
			helpDialog = new HelpDialog(this);
		helpDialog.show();
	}

	private void doAbout() {
		if (about == null)
			setAboutPanel();
		JOptionPane.showMessageDialog(this, about, "ABOUT",
				JOptionPane.PLAIN_MESSAGE, new ImageIcon(
						loadImage("jetris.png")));
	}

	private void setAboutPanel() {
		about = new JPanel();
		about.setLayout(new BoxLayout(about, BoxLayout.Y_AXIS));
		JLabel jl = new JLabel("<HTML><B>" + NAME
				+ "</B> Copyright (c) 2006 Nikolay G. Georgiev</HTML>");
		jl.setFont(font);
		about.add(jl);
		about.add(Box.createVerticalStrut(10));

		jl = new JLabel("WEB PAGE:");
		jl.setFont(font);
		about.add(jl);
		HTMLLink hl = new HTMLLink("http://jetris.sf.net", false);
		hl.setFont(font);
		about.add(hl);

		about.add(Box.createVerticalStrut(20));

		jl = new JLabel(
				"<HTML>This program is released under the Mozilla Public License 1.1 .<BR> A copy of this is included with your copy of JETRIS<BR>and can also be found at:</HTML>");
		jl.setFont(font);
		about.add(jl);
		about.add(jl);
		hl = new HTMLLink("http://www.opensource.org/licenses/mozilla1.1.php",
				false);
		hl.setFont(font);
		about.add(hl);
	}

	private void showHiScore() {
		setHiScorePanel();

		JOptionPane.showMessageDialog(this, hiScorePanel, "HI SCORE",
				JOptionPane.PLAIN_MESSAGE, new ImageIcon(
						loadImage("jetris32x32.png")));

		hiScorePanel = null;
	}

	private void setHiScorePanel() {
		hiScorePanel = new JPanel(new BorderLayout());

		String[] colNames = { "Place", "Points", "Lines", "Name" };
		String[][] data = new String[tg.hiScore.length + 1][colNames.length];
		data[0] = colNames;
		for (int i = 0; i < tg.hiScore.length; i++) {
			data[i + 1] = new String[colNames.length];
			data[i + 1][0] = (i + 1) + ".";
			data[i + 1][1] = ("" + tg.hiScore[i].score);
			data[i + 1][2] = ("" + tg.hiScore[i].lines);
			data[i + 1][3] = ("" + tg.hiScore[i].name);
		}

		JTable table = new JTable(data, colNames);
		table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
		table.setBackground(new Color(230, 255, 255));
		table.setEnabled(false);

		hiScorePanel.add(table, BorderLayout.CENTER);
		JButton jb = new JButton("Publish HiScore Online");
		jb.addActionListener(pH);

		hiScorePanel.add(jb, BorderLayout.SOUTH);
	}

	private class PublishHandler implements ActionListener {
		public void actionPerformed(ActionEvent ae) {
			JButton jb = (JButton) ae.getSource();
			PublishThread pt = new PublishThread(jb);
			pt.start();
		}
	}

	private class PublishThread extends Thread {

		private JButton but;

		PublishThread(JButton source) {
			super();
			but = source;
		}

		public void run() {
			but.setEnabled(false);
			boolean b = false;
			try {
				for (int i = 0; i < tg.hiScore.length; i++) {
					PublishHiScore.publish(tg.hiScore[i]);
				}
			} catch (Exception e) {
				e.printStackTrace();
				b = true;
				JOptionPane.showMessageDialog(hiScorePanel,
						"Could not publish HiScore online!\nTry again later!",
						"ERROR", JOptionPane.ERROR_MESSAGE);
			}
			if (!b) {
				JOptionPane.showMessageDialog(hiScorePanel,
						"Publishing HiScore was successfull :)", "HI SCORE",
						JOptionPane.INFORMATION_MESSAGE);
			}
			but.setEnabled(true);
		}

	}

	private class MenuHandler implements ActionListener {

		public void actionPerformed(ActionEvent e) {
			try {
				JMenuItem tmp = (JMenuItem) e.getSource();
				if (tmp == jetrisRestart) {
					restart();
				} else if (tmp == jetrisPause) {
					pause();
				} else if (tmp == jetrisHiScore) {
					showHiScore();
				} else if (tmp == jetrisExit) {
					System.exit(0);
				} else if (tmp == helpJetris) {
					doHelp();
				} else if (tmp == helpAbout) {
					doAbout();
				}
			} catch (Exception exc) {
				exc.printStackTrace(System.out);
			}
		}
	}
}
